"""
Pykrete versioning.version tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
from pykrete.versioning import RevisionType, Version
from .versioning import PykreteVersioningTestCase


class PykreteVersioningVersionTestCase(PykreteVersioningTestCase):
    """Unit tests for pykrete's versioning module's version class
    """
    def test_version_from_tuple(self):
        """Verifies creation of version from tuple"""
        target = Version((1, 2, 3, 4, RevisionType.Alpha))
        self._assert_version(target, (1, 2, 3, 4, RevisionType.Alpha))

    def test_version_from_args(self):
        """Verifies creation of version from arguments"""
        target = Version(12, 23, 34, 45, RevisionType.Beta)
        self._assert_version(target, (12, 23, 34, 45, RevisionType.Beta))

    def test_version_from_version(self):
        """Verifies creation of version from another version object"""
        source = Version(111, 222, 333, 444, RevisionType.RC)
        target = Version(source)
        self._assert_version(target, (111, 222, 333, 444, RevisionType.RC))
        self.assertEqual(source, target, "Version objects aren't equal")

    def test_version_from_kwargs(self):
        """Verifies creation of version from named arguments"""
        target = Version(major=21, minor=32, revision=43, revision_type=RevisionType.Release,
                         build=54)
        self._assert_version(target, (21, 32, 43, 54, RevisionType.Release))

    def test_version_ordering(self):
        """Verifies ordering of version objects"""
        source = [Version(1, 2, 3, 4, RevisionType.Release),
                  Version(1, 2, 3, 1, RevisionType.Release),
                  Version(1, 2, 3, 4, RevisionType.Release),
                  Version(1, 2, 1, 3, RevisionType.Release),
                  Version(1, 0, 2, 3, RevisionType.Release),
                  Version(0, 1, 2, 3, RevisionType.Release),
                  Version(11, 22, 33, 44, RevisionType.Release)]
        target = sorted(source)
        print('\n'.join(str(version) for version in target))
        expected = [Version(0, 1, 2, 3, RevisionType.Release),
                    Version(1, 0, 2, 3, RevisionType.Release),
                    Version(1, 2, 1, 3, RevisionType.Release),
                    Version(1, 2, 3, 1, RevisionType.Release),
                    Version(1, 2, 3, 4, RevisionType.Release),
                    Version(1, 2, 3, 4, RevisionType.Release),
                    Version(11, 22, 33, 44, RevisionType.Release)]
        for comparison in zip(expected, target):
            self.assertEqual(comparison[0], comparison[1], "Wrong ordering")

    def _assert_version(self, target, expected):
        (major, minor, revision, build, revision_type) = expected
        print(target)
        self._assert_version_pattern(target)
        self.assertEqual(major, target.major, "Wrong major version")
        self.assertEqual(minor, target.minor, "Wrong minor version")
        self.assertEqual(revision, target.revision, "Wrong revision")
        self.assertEqual(build, target.build, "Wrong build number")
        self.assertEqual(revision_type, target.revision_type, "Wrong revision type")


if __name__ == '__main__':
    unittest.main()



"""
SRC path appender for local use of pykrete
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import sys
from os.path import dirname


def print_path():
    """prints sys.path"""
    print(str(sys.path))


sys.path.append(dirname(__file__) + '\\src')
print_path()



"""
GIT repository management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
from git import Repo


class Git:
    """Handles a GIT repository"""

    def __init__(self, path='.'):
        self._repo = Repo.init(path)

    def get_all_tags(self):
        """Gets all tags from the current branch in the repo

        :return: A list of tag info tuples (name, message)
        """
        return self._get_tags()

    def get_tags_from(self, pattern, is_must=False):
        """Gets all tags from the current branch in the repo, from the tag whose name matches the
         supplied pattern.

        :param pattern: name pattern to match
        :param is_must: True to raise an error if no match is found, False to return all tags in
        that case.
        :return: A list of tag info tuples (name, message)
        """
        return self._get_tags(lambda tag_refs: self.__filter_tags_from(pattern, is_must, tag_refs))

    def _get_tags(self, tag_ref_filter=None):
        tag_refs = list(self._repo.tags)
        requested_tag_references = tag_ref_filter(tag_refs) if tag_ref_filter else tag_refs
        return self.__tag_refs_to_name_message_tuples(requested_tag_references)

    def __filter_tags_from(self, pattern, is_must, tag_refs):
        from_index = self.__find_tag_index(pattern, tag_refs)
        if is_must and from_index == -1:
            raise KeyError("No tag found to match pattern " + pattern)
        return tag_refs[from_index if from_index > 0 else 0:]

    @staticmethod
    def __find_tag_index(pattern, tag_refs):
        compiled_pattern = re.compile(pattern)
        for i in range(len(tag_refs) - 1, -1, -1):
            if compiled_pattern.match(tag_refs[i].name):
                return i
        return -1

    @staticmethod
    def __tag_refs_to_name_message_tuples(tags):
        return [(tag.name, tag.tag.message if tag.tag else None) for tag in tags]



"""
Python package
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from re import findall
from os import path
from pkg_resources import require


def make_python_root_version_file_path(project):
    """ Generate the version.py file from the project's name

    :param project: project name
    :return: src/<project>/version.py
    """
    return path.join('src', project, 'version.py')


class PythonPackage:
    """Python package"""

    @property
    def project(self):
        """
        :return: (string) name of project package under 'src'.
        """
        return self._project

    @property
    def version(self):
        """
        :return: (string) The project's version
        """
        return self._version

    def __init__(self, project, package_fallback=False):
        """Initializes this instance to analyze the specified project

        :param project: name of project package under 'src'
        """
        self._version_file = make_python_root_version_file_path(project)
        self._project = project
        self._package_fallback = package_fallback
        self._version = self._get_version()

    @staticmethod
    def get_long_description():
        """Gets the contents of the README.md file

        :return: Long description
        """
        with open('README.md', 'r') as readme:
            return readme.read()

    def _get_version(self):
        """Gets the package version

        :return: The version
        :exception IndexError: Version not found
        """
        try:
            return self._get_version_from_file()
        except IOError:
            if not self._package_fallback:
                raise
            version_from_current_package = require(self._project)[0].version
            return version_from_current_package

    def _get_version_from_file(self):
        """Gets the version from the version file

        :return: The version
        :exception IOError: Version file doesn't exist
        """
        try:
            with open(self._version_file, 'r') as file:
                return findall("__version__ = '([^']*)'", file.read())[0]
        except IndexError:
            raise IOError('Version not found in ' + self._version_file)

    def __str__(self):
        return f'{self._project} v{self._version}'



"""
CI environment-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from pykrete.args import environ
from .revision_type import RevisionType
from .version import Version


class CiVersion(Version):
    """Handle version in CI environment
    Revision is:
        Release - in master branch
        RC - in non-WIP merge request
        Beta - in merge request with 'WIP' in the title
        Alpha - none of the above
    """
    def __init__(self, ci_reader=environ, ci_spec=None):
        """Initializes this instance from CI environment

        :param ci_reader: (ci_variable, role)=>value function [see pykrete.args.environ behavior]
        :param ci_spec: dictionary of part to CI variable
         (parts are: 'major version', 'minor version', 'build version', 'branch name',
          'merge request title')
        """
        self.__ci_reader = ci_reader
        self.__ci_spec = ci_spec if ci_spec else self.__get_default_ci_spec()
        revision_type = self.__get_revision_type()
        super().__init__(major=self.__read_env('major version'),
                         minor=self.__read_env('minor version'),
                         revision=revision_type.value,
                         build=self.__read_env('build version'),
                         revision_type=revision_type)

    def __get_revision_type(self):
        if self.__is_master():
            return RevisionType.Release

        revision = self.__revision_from_merge_request()
        return revision if revision else RevisionType.Alpha

    def __read_env(self, part, is_no_role=False):
        return self.__ci_reader(self.__ci_spec[part],
                                None if is_no_role else f'CI\'s \'{part}\' value')

    def __is_master(self):
        branch_name = self.__read_env('branch name')
        return branch_name == 'master'

    def __revision_from_merge_request(self):
        merge_request_title = self.__read_env('merge request title', True)
        if merge_request_title:
            return RevisionType.Beta if 'WIP' in merge_request_title else RevisionType.RC
        return None

    @staticmethod
    def __get_default_ci_spec():
        return {'major version': 'CI_VERSION_MAJOR',
                'minor version': 'CI_VERSION_MINOR',
                'build version': 'CI_PIPELINE_IID',
                'branch name': 'CI_COMMIT_REF_NAME',
                'merge request title': 'CI_MERGE_REQUEST_TITLE'}



"""
Pykrete root
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .version import __version__

__all__ = ['__version__', 'args', 'builder', 'calls', 'distrib', 'logging', 'packages',
           'versioning']



"""
Pykrete logging
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .make_logger import make_logger, make_logger_by_verbose_arg
from .fatal import fatal

__all__ = ['make_logger', 'make_logger_by_verbose_arg', 'fatal']



"""
version.py version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from re import findall
from pykrete.packages import make_python_root_version_file_path
from .version import Version
from .revision_type import RevisionType


class VersionPyVersion(Version):
    """Handle version in python.py"""
    revisions = {
        '.': RevisionType.Release,
        'rc': RevisionType.RC,
        'b': RevisionType.Beta,
        'a': RevisionType.Alpha}

    def __init__(self, project):
        """Initialize this instance from the specified project's version.py file

        :param project: Python project (under 'src')
        """
        version_py_path = make_python_root_version_file_path(project)
        version_match = self.__read_version_match(version_py_path)
        version_parts = self.__parse_version_parts(version_match)
        super().__init__(version_parts)

    @staticmethod
    def __read_version_match(version_py_path):
        try:
            with open(version_py_path, 'r') as file:
                separator_pattern = '|'.join(VersionPyVersion.revisions.keys()).replace('.', r'\.')
                return findall(f"__version__ = '(\\d+)\\.(\\d+)({separator_pattern})(\\d+)'",
                               file.read())[0]
        except IndexError:
            raise IOError("No version specification found in file")

    @staticmethod
    def __parse_version_parts(version_parts):
        if not version_parts or len(version_parts) != 4:
            raise IOError('Incompatible version')
        revision_marker = version_parts[2]
        revision_type = VersionPyVersion.revisions[revision_marker]
        return (version_parts[0], version_parts[1], revision_type.value, version_parts[3],
                revision_type)



"""
Pykrete packages tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
from pykrete.packages import PythonPackage
from .helpers import get_full_path


class PythonPackageTestCase(unittest.TestCase):
    """Unit tests for pykrete's args module
    """
    def test_release_version_good(self):
        """Verifies reading of a release package's version
        """
        self._assert_good_package('african_swallow', '12.34.56')

    def test_rc_version_good(self):
        """Verifies reading of a release-candidate package's version
        """
        self._assert_good_package('european_swallow', '12.34rc56')

    def test_missing_version_in_file(self):
        """Verifies handling of a bad version file
        """
        with self.assertRaises(IOError):
            print(str(PythonPackage(get_full_path('a_duck'))))

    def test_missing_version_file(self):
        """Verifies handling of a missing version file
        """
        with self.assertRaises(FileNotFoundError):
            print(str(PythonPackage(get_full_path('a_witch'))))

    def _assert_good_package(self, project, expected_version):
        """Assert version as expected"""
        target = PythonPackage(get_full_path(project))
        self.assertEqual(expected_version, target.version, 'Wrong version read')


if __name__ == '__main__':
    unittest.main()



"""
Python build info
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import os
from pykrete.args import environ
from .build_info import BuildInfo


class PythonBuildInfo(BuildInfo):
    """Python-specific build info"""
    # pylint: disable=too-few-public-methods
    def make_version(self):
        """Sets the package version from GitLab environment variables:
        CI_VERSION_MAJOR.CI_VERSION_MINOR.CI_PIPELINE_IID
        On any branch other than maser, CI_PIPELINE_ID is preceded by 'rc' instead of '.'

        :return: The version string
        """
        return ''.join(
            [
                environ('CI_VERSION_MAJOR', 'the major version of the package'),
                '.',
                environ('CI_VERSION_MINOR', 'the minor version of the package'),
                '.' if self.release else 'rc',
                os.environ['CI_PIPELINE_IID']
            ])



"""
Pykrete logger tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import unittest

from pykrete.logging import make_logger


class PykreteLoggerUnitTests(unittest.TestCase):
    """Test pykrete's logger module"""
    def test_logger_default_stdout(self):
        """Verifies that only info is sent to stdout
        """
        logger = make_logger('default_stdout', False)
        logger.info('You should see this info')
        logger.debug('You shouldn\'t see this debug')
        logger.info(self.longMessage)  # how can I test this?

    def test_logger_verbose_stdout(self):
        """Verifies that both debug and info are sent to stdout when verbose output is requested
        """
        logger = make_logger('default_stdout', True)
        logger.info('You should see this info')
        logger.debug('You should see this debug as well')
        logger.info(self.longMessage)  # how can I test this?

    def test_logger_stderr(self):
        """Verifies that only info is sent to stdout
        """
        logger = make_logger('default_stdout', False)
        logger.error('You should see this error')
        logger.critical('You should see this critical')
        logger.exception('You should see this exception')
        logger.info(self.longMessage)  # how can I test this?


if __name__ == '__main__':
    unittest.main()



"""
Pykrete versioning.version_py_version tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
import os
from pykrete.versioning import VersionPyVersion
from .versioning import PykreteVersioningTestCase


def _get_full_path(project):
    return os.path.join('pykrete', 'tests', 'src', project)


class PykreteVersioningVersionPyVersionTestCase(PykreteVersioningTestCase):
    """Unit tests for pykrete's versioning module's VersionPyVersion class
    """
    def test_version_py_release_version_good(self):
        """Verifies reading of a release package's version
        """
        self._assert_good_package('african_swallow', '12.34.4.56 Release')

    def test_version_py_rc_version_good(self):
        """Verifies reading of a release-candidate package's version
        """
        self._assert_good_package('european_swallow', '12.34.3.56 RC')

    def test_version_py_missing_version_in_file(self):
        """Verifies handling of a bad version file
        """
        with self.assertRaises(IOError):
            print(str(VersionPyVersion(_get_full_path('a_duck'))))

    def test_version_py_missing_version_file(self):
        """Verifies handling of a missing version file
        """
        with self.assertRaises(FileNotFoundError):
            print(str(VersionPyVersion(_get_full_path('a_witch'))))

    def _assert_good_package(self, project, expected_version):
        """Assert version as expected"""
        target = VersionPyVersion(_get_full_path(project))
        print(target)
        self._assert_version_pattern(target)
        self.assertEqual(expected_version, str(target), 'Wrong version read')


if __name__ == '__main__':
    unittest.main()



"""
Python package installation
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import subprocess


def install_python_packages(logger, *args):
    """Installs the specified pip packages

    :param logger: Optional logger for debug print(s)
    :param args: The packages to install
    """
    packages = list(args)
    if logger:
        logger.debug('Installing packages %s', str(packages))
    subprocess.check_call(['python', '-m', 'pip', 'install'] + packages)



"""
Pykrete logger tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import sys
import os
from abc import abstractmethod
from pykrete.logging import make_logger_by_verbose_arg, fatal


class BuildInfo:
    """Represents the selected build mode"""

    @property
    def release(self):
        """
        :return: (bool) True for Release mode, False otherwise
        """
        return self._release

    @property
    def upload(self):
        """
        :return: (bool) True for Upload of build, False otherwise
        """
        return self._upload

    def __init__(self):
        self._logger = make_logger_by_verbose_arg('BuildMode')
        self._release = self._is_release()
        self._upload = self._release or 'candidate' in sys.argv
        self._forced = self._is_forced()
        self._upload |= self._forced

    @abstractmethod
    def make_version(self):
        """Makes the version string for this build

        :return: The version string
        """

    @staticmethod
    def _is_release():
        """Checks whether this is a Release build

        :return: True for release build, False otherwise
        :exception Exception: Release mode requested outside of the master branch
        """
        release = 'release' in sys.argv
        if release and not os.environ['CI_COMMIT_REF_NAME'] == 'master':
            fatal('Only master branch can run release build')
        return release

    def _is_forced(self):
        """Checks if upload is manually forced

        :return: True if manually forced, False otherwise
        :exception FileExistsError: system forced the build, but the manual force still exists
        """
        marker_file = 'FORCE_UPLOAD.txt'
        self._logger.debug('Checking for %s', os.path.join(os.getcwd(), marker_file))
        file_exists = os.path.isfile(marker_file)
        if file_exists and self._upload:
            fatal(f'Remove manual commit file [{marker_file}]', self._logger, FileExistsError)
        return file_exists

    def __str__(self):
        """Gets a string representation of this object"""
        return ''.join([
            'Release' if self._release else 'Candidate'
            ' build with',
            ' forced' if self._forced else ('' if self._upload else 'out'),
            ' upload of version ',
            self.make_version()])



"""
Argument type verification
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import os
import argparse
import re


def _string_to_version(arg_string):
    """Convert a string to a match with parts as groups

    :param arg_string: version string
    :return: match object with version parts as groups
    """
    return re.match(r'^(\d+).(\d+).(\d+).(\d+)$', arg_string)


def build_version(arg_string):
    """Argument verification for valid build versions

    :param arg_string: version string
    :return: match object with version parts as groups
    :exception: invalid version string
    """
    version = _string_to_version(arg_string)
    if version:
        return version
    raise argparse.ArgumentTypeError(f'Invalid version: {arg_string}')


def exiting_file(arg_string):
    """Argument verification for existing-file paths

    :param arg_string: full or relative file path
    :return: the path, as received
    :exception: illegal path or file doesn't exist
    """
    if arg_string and os.path.isfile(arg_string):
        return arg_string
    raise argparse.ArgumentTypeError(f'File {arg_string} not found')


def yes_or_no(arg_string):
    """Argument verification for 'yes'/'no' flags

    :param arg_string: parameter string
    :return: True of 'yes', false if 'no'
    :exception: other values
    """
    answer = arg_string.lower().strip()
    if answer == 'yes':
        return True
    if answer == 'no':
        return False
    raise argparse.ArgumentTypeError(f'Expected a yes/no value, got "{answer}"')



"""
Pykrete repo tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
from pykrete.repo import Git
from .helpers import get_full_path


class PythonRepoTestCase(unittest.TestCase):
    """Unit tests for pykrete's repo module
    """
    def test_get_all_tags(self):
        """Verifies getting all tags from repo
        """
        self._assert_get_tags(lambda git: git.get_all_tags(),
                              ('1.0.0', None), ('1.0.1', 'bug fix'))

    def test_get_tags_from_existing(self):
        """Verifies getting all tags from repo starting from a specific tag
        """
        self._assert_get_tags(lambda git: git.get_tags_from(r'.*?0\.1', True),
                              ('1.0.1', 'bug fix'))

    def test_get_tags_from_must_but_non_existing(self):
        """Verifies getting all tags from repo
        """
        with self.assertRaises(KeyError):
            self._assert_get_tags(lambda git: git.get_tags_from(r'no such tag', True))

    def test_get_tags_from_non_must_non_existing(self):
        """Verifies getting all tags from repo
        """
        self._assert_get_tags(lambda git: git.get_tags_from(r'no such tag', False),
                              ('1.0.0', None), ('1.0.1', 'bug fix'))

    def _assert_get_tags(self, getter, *expected):
        """Verifies tags from repo

        :param getter: tag getter (lambda on Git object)
        :param expected: expected returned tags
        :return:
        """
        test_repo_path = get_full_path('git-tag-example', True)
        target = Git(test_repo_path)
        tags = getter(target)
        print(tags)
        self.assertTrue(tags, 'Got empty tags')
        self.assertTrue(len(tags) >= len(expected), 'too few tags returned')
        for test in zip(expected, tags):
            self.assertEqual(test[0], test[1], 'wrong tag returned')


if __name__ == '__main__':
    unittest.main()



"""
Pykrete calls tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import unittest

from pykrete.calls import CheckedCall


class PykreteCallsUnitTests(unittest.TestCase):
    """Test pykrete's call module"""
    def test_checked_command_success(self):
        """Verifies successful call and correct passing of normal and quoted parameters
        with different quote types
        """
        for is_vector in [False, True]:
            target = self._make_checked_call_to_listargs_cmd_with(
                ['1', 'two', 'three and some', 'four',
                 'middle',
                 'four', 'three and some', 'two', '1',
                 '1 two "three and some" \'four\' middle four \'three and some\' \'two\' "1"'
                 ],
                is_vector)
            print(str(target))
            self.__assert_true(target.success, 'command failed', is_vector)
            self.__assert_true('PASSED' in target.stdout, 'stdout did not contain PASSED',
                               is_vector)

    def test_checked_command_failure(self):
        """Verifies failing call and correct passing of arguments with mixed quotes
        """
        args = ['start "middle" end', '2',
                '3', '4', '5', '6', '7', '8', 'start \'middle\' end',
                '\'start "middle" end\' 2 3 4 5 6 7 8 "start \'middle\' end"']
        for is_vector in [False, True]:
            target = self._make_checked_call_to_listargs_cmd_with(args, is_vector)
            print(str(target))
            self.__assert_true(not target.success, 'command succeeded', is_vector)
            self.__assert_true('FAILED' in target.stderr, 'stderr did not contain FAILED',
                               is_vector)

    @staticmethod
    def _make_checked_call_to_listargs_cmd_with(params, is_vector):
        listargs = r'src\pykrete\tests\src\listargs.cmd'
        if is_vector:
            cmd = [listargs]
            cmd.extend(params[:-1])
        else:
            cmd = f'{listargs} {params[9]}'
        return CheckedCall(cmd)

    def __assert_true(self, expr, message, is_vector):
        self.assertTrue(expr, f'{message} [{"vector" if is_vector else "string"} call]')


if __name__ == '__main__':
    unittest.main()



"""
Build manager base class
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import abstractmethod
from pykrete.logging import make_logger_by_verbose_arg
from pykrete.args import environ
from pykrete.builder.info.build_info import BuildInfo


class BuildManager:
    """Manages building a python package"""

    def __init__(self, build_info: BuildInfo):
        """Initializes this instance
        :type build_info: Build info
        """
        self._info = build_info
        self._logger = make_logger_by_verbose_arg('BuildManager')
        self._project = environ('CI_TARGET', 'the name of the project under src')

    def run(self):
        """Runs the build process"""
        self._clean_distribution()
        self._set_package_version()
        self._build()
        if not self._info.upload:
            self._logger.info('Upload skipped')
            return
        self._upload()

    @abstractmethod
    def _clean_distribution(self):
        """Removes previous distribution files"""

    @abstractmethod
    def _set_package_version(self):
        """Sets the package version in the project's version file"""

    @abstractmethod
    def _build(self):
        """Performs the actual build using setup.py"""

    @abstractmethod
    def _upload(self):
        """Uploads the package"""

    def __str__(self):
        """Gets a string representation of this object"""
        return f'{self._info} of project "{self._project}"'



"""
StdOut and StdErr log maker
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
import sys


class _LessThanFilter(logging.Filter):
    """Logging filter for messages below a certain level
    """

    def __init__(self, exclusive_maximum, name=""):
        """Initialize this instance to filter messages below the maximum

        :param exclusive_maximum: Exclusive maximum log level
        :param name: filter name (optional)
        """
        super(_LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        """Filter log records

        :param record: log record
        :return: 1 if record is to be shown, 0 otherwise
        """
        return 1 if record.levelno < self.max_level else 0

    def __str__(self):
        return "LessThan " + self.max_level


def make_logger_by_verbose_arg(name):
    """Makes a logger which prints info into stdout, and all others to stderr
    If '--verbose' appears in the scrip arguments, debug is also sent to stdout

    :param name: logger name
    :return: logger
    """
    return make_logger(name, '--verbose' in sys.argv)


def make_logger(name, is_verbose):
    """Makes a logger which prints info into stdout, and all others to stderr
    If isVerbose' is set, debug is also sent to stdout

    :param is_verbose: Should debug appear in log
    :param name: logger name
    :return: logger
    """
    new_logger = logging.getLogger(name)
    if is_verbose:
        new_logger.setLevel(logging.DEBUG)
    else:
        new_logger.setLevel(logging.INFO)

    logging_handler_out = logging.StreamHandler(sys.stdout)
    logging_handler_out.setLevel(logging.DEBUG)
    logging_handler_out.addFilter(_LessThanFilter(logging.WARNING))
    new_logger.addHandler(logging_handler_out)

    logging_handler_err = logging.StreamHandler(sys.stderr)
    logging_handler_err.setLevel(logging.WARNING)
    new_logger.addHandler(logging_handler_err)
    return new_logger



"""
Running parse context
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


class ParseContext:
    """keeper of the bridge..."""
    def __init__(self):
        self.last_char = None
        self.__vector = []
        self.__current_arg = []

    def append_last_char_to_current_arg(self):
        """another char in the wall"""
        self.__current_arg.append(self.last_char)

    def flush_current_arg(self):
        """prep for next arg"""
        self._flush_current_arg_to_vector(self.__vector)
        self.__current_arg = []

    def get_vector(self):
        """
        :return: collected vector
        """
        vector = self.__vector.copy()
        self._flush_current_arg_to_vector(vector)
        return vector

    def _flush_current_arg_to_vector(self, vector):
        if self.__current_arg:
            vector.append(''.join(self.__current_arg[:-1]))

    def __str__(self):
        return f'{self.__vector}'



"""
Command-to-vector state machine
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .get_starting_state import get_starting_state
from .parsing_state import ParsingState

__all__ = ['get_starting_state', 'ParsingState']



"""
Parse command into call vector
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .parse_command_into_call_vector_state_machine import get_starting_state


def parse_command_into_call_vector(cmd):
    """Parse command into a call-vector

    :param cmd: The command
    :return: The call vector
    """
    return _run_state_machine_loop_and_get_vector_from_end_state(cmd)


def _run_state_machine_loop_and_get_vector_from_end_state(cmd):
    state = get_starting_state()
    for char in cmd:
        state = state.parse_and_get_next_state(char)
    return state.parse_and_get_next_state(' ').get_vector()



"""
Base class for parsing states
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import abstractmethod

from .parse_context import ParseContext


class ParsingState:
    """Parsing state
    """
    _whitespace_chars = [' ', '\t']
    _quote_chars = ['\'', '"']

    def __init__(self, context: ParseContext):
        self._context = context

    def parse_and_get_next_state(self, current_char):
        """
        :param current_char: parsed character
        :return: next state
        """
        self._on_new_char(current_char)
        if self.__is_known_char(current_char):
            return self.__get_known_next_step(current_char)
        return self.__get_real_default_step()

    def get_vector(self):
        """
        :return: collected vector
        """
        return self._context.get_vector()

    def __str__(self):
        return f'State is {self.__class__.__name__}'

    def _on_new_char(self, char):
        self._context.last_char = char

    @abstractmethod
    def _get_default_next_state(self):
        """
        :return: either a state-class type, which will be used to create the default next step
         (if no char in the below dict is matched), or None which means the current state remains
          in effect
        """

    @abstractmethod
    def _get_dict_of_chars_to_new_states(self):
        """
        :return: dict where the values are wither a state-class type, which will be used to create
         the next step is that key is matched, or None which means the current state will remain in
          effect
        """

    def __is_known_char(self, char):
        return char in self._get_dict_of_chars_to_new_states().keys()

    def __get_known_next_step(self, char):
        next_state = self._get_dict_of_chars_to_new_states()[char]
        return self.__get_real_next_step(next_state)

    def __get_real_default_step(self):
        next_state = self._get_default_next_state()
        return self.__get_real_next_step(next_state)

    def __get_real_next_step(self, next_step):
        return next_step(self._context) if next_step else self
